import jax
from functools import partial

from craftax.craftax.constants import *
from craftax.craftax.craftax_state import EnvState
from craftax.craftax.util.game_logic_utils import is_boss_vulnerable


def render_craftax_symbolic(state: EnvState):
    map = state.map[state.player_level]

    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)

    # Map
    padded_grid = jnp.pad(
        map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )

    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2

    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)
    map_view_one_hot = jax.nn.one_hot(map_view, num_classes=len(BlockType))

    # Items
    padded_items_map = jnp.pad(
        state.item_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=ItemType.NONE.value,
    )

    item_map_view = jax.lax.dynamic_slice(padded_items_map, tl_corner, OBS_DIM)
    item_map_view_one_hot = jax.nn.one_hot(item_map_view, num_classes=len(ItemType))

    # Mobs
    mob_types_per_class = 8
    mob_map = jnp.zeros(
        (*OBS_DIM, 5 * mob_types_per_class), dtype=jnp.int32
    )  # 5 classes * 8 types

    def _add_mob_to_map(carry, mob_index):
        mob_map, mobs, mob_class_index = carry

        local_position = (
            mobs.position[mob_index]
            - state.player_position
            + jnp.array([OBS_DIM[0], OBS_DIM[1]]) // 2
        )
        on_screen = jnp.logical_and(
            local_position >= 0, local_position < jnp.array([OBS_DIM[0], OBS_DIM[1]])
        ).all()
        on_screen *= mobs.mask[mob_index]

        mob_identifier = mob_class_index * mob_types_per_class + mobs.type_id[mob_index]
        mob_map = mob_map.at[local_position[0], local_position[1], mob_identifier].set(
            on_screen.astype(jnp.int32)
        )

        return (mob_map, mobs, mob_class_index), None

    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.melee_mobs), 0),
        jnp.arange(state.melee_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.passive_mobs), 1),
        jnp.arange(state.passive_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.ranged_mobs), 2),
        jnp.arange(state.ranged_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (
            mob_map,
            jax.tree_map(lambda x: x[state.player_level], state.mob_projectiles),
            3,
        ),
        jnp.arange(state.mob_projectiles.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (
            mob_map,
            jax.tree_map(lambda x: x[state.player_level], state.player_projectiles),
            4,
        ),
        jnp.arange(state.player_projectiles.mask.shape[1]),
    )

    all_map = jnp.concatenate(
        [map_view_one_hot, item_map_view_one_hot, mob_map], axis=-1
    )

    # Light map
    padded_light_map = jnp.pad(
        state.light_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=0.0,
    )
    light_map_view = jax.lax.dynamic_slice(padded_light_map, tl_corner, OBS_DIM) > 0.05

    # Mask out tiles and mobs in darkness
    all_map = all_map * light_map_view[:, :, None]
    all_map = jnp.concatenate(
        (all_map, jnp.expand_dims(light_map_view, axis=-1)), axis=-1
    )

    # Inventory
    inventory = jnp.array(
        [
            jnp.sqrt(state.inventory.wood) / 10.0,
            jnp.sqrt(state.inventory.stone) / 10.0,
            jnp.sqrt(state.inventory.coal) / 10.0,
            jnp.sqrt(state.inventory.iron) / 10.0,
            jnp.sqrt(state.inventory.diamond) / 10.0,
            jnp.sqrt(state.inventory.sapphire) / 10.0,
            jnp.sqrt(state.inventory.ruby) / 10.0,
            jnp.sqrt(state.inventory.sapling) / 10.0,
            jnp.sqrt(state.inventory.torches) / 10.0,
            jnp.sqrt(state.inventory.arrows) / 10.0,
            state.inventory.books / 2.0,
            state.inventory.pickaxe / 4.0,
            state.inventory.sword / 4.0,
            state.sword_enchantment,
            state.bow_enchantment,
            state.inventory.bow,
        ]
    ).astype(jnp.float32)

    potions = jnp.sqrt(state.inventory.potions) / 10.0
    armour = state.inventory.armour / 2.0
    armour_enchantments = state.armour_enchantments

    intrinsics = jnp.array(
        [
            state.player_health / 10.0,
            state.player_food / 10.0,
            state.player_drink / 10.0,
            state.player_energy / 10.0,
            state.player_mana / 10.0,
            state.player_xp / 10.0,
            state.player_dexterity / 10.0,
            state.player_strength / 10.0,
            state.player_intelligence / 10.0,
        ]
    ).astype(jnp.float32)

    direction = jax.nn.one_hot(state.player_direction - 1, num_classes=4)

    special_values = jnp.array(
        [
            state.light_level,
            state.is_sleeping,
            state.is_resting,
            state.learned_spells[0],
            state.learned_spells[1],
            state.player_level / 10.0,
            state.monsters_killed[state.player_level] >= MONSTERS_KILLED_TO_CLEAR_LEVEL,
            is_boss_vulnerable(state),
        ]
    )

    all_flattened = jnp.concatenate(
        [
            all_map.flatten(),
            inventory,
            potions,
            intrinsics,
            direction,
            armour,
            armour_enchantments,
            special_values,
        ]
    )

    return all_flattened


@partial(
    jax.jit,
    static_argnums=(
        1,
        2,
    ),
)
def render_craftax_pixels(state, block_pixel_size, do_night_noise=True):
    textures = TEXTURES[block_pixel_size]
    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)

    # RENDER MAP
    # Get view of map
    map = state.map[state.player_level]
    padded_grid = jnp.pad(
        map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )

    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2

    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)

    # Boss
    boss_block = jax.lax.select(
        is_boss_vulnerable(state),
        BlockType.NECROMANCER_VULNERABLE.value,
        BlockType.NECROMANCER.value,
    )

    map_view_boss = map_view == BlockType.NECROMANCER.value
    map_view = map_view_boss * boss_block + (1 - map_view_boss) * map_view

    # Render map tiles
    map_pixels_indexes = jnp.repeat(
        jnp.repeat(map_view, repeats=block_pixel_size, axis=0),
        repeats=block_pixel_size,
        axis=1,
    )
    map_pixels_indexes = jnp.expand_dims(map_pixels_indexes, axis=-1)
    map_pixels_indexes = jnp.repeat(map_pixels_indexes, repeats=3, axis=2)

    map_pixels = jnp.zeros(
        (OBS_DIM[0] * block_pixel_size, OBS_DIM[1] * block_pixel_size, 3),
        dtype=jnp.float32,
    )

    def _add_block_type_to_pixels(pixels, block_index):
        return (
            pixels
            + textures["full_map_block_textures"][block_index]
            * (map_pixels_indexes == block_index),
            None,
        )

    map_pixels, _ = jax.lax.scan(
        _add_block_type_to_pixels, map_pixels, jnp.arange(len(BlockType))
    )

    # Items
    padded_item_map = jnp.pad(
        state.item_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=ItemType.NONE.value,
    )

    item_map_view = jax.lax.dynamic_slice(padded_item_map, tl_corner, OBS_DIM)

    # Insert blocked ladders
    is_ladder_down_open = (
        state.monsters_killed[state.player_level] >= MONSTERS_KILLED_TO_CLEAR_LEVEL
    )
    ladder_down_item = jax.lax.select(
        is_ladder_down_open,
        ItemType.LADDER_DOWN.value,
        ItemType.LADDER_DOWN_BLOCKED.value,
    )

    item_map_view_is_ladder_down = item_map_view == ItemType.LADDER_DOWN.value
    item_map_view = (
        item_map_view_is_ladder_down * ladder_down_item
        + (1 - item_map_view_is_ladder_down) * item_map_view
    )

    map_pixels_item_indexes = jnp.repeat(
        jnp.repeat(item_map_view, repeats=block_pixel_size, axis=0),
        repeats=block_pixel_size,
        axis=1,
    )
    map_pixels_item_indexes = jnp.expand_dims(map_pixels_item_indexes, axis=-1)
    map_pixels_item_indexes = jnp.repeat(map_pixels_item_indexes, repeats=3, axis=2)

    def _add_item_type_to_pixels(pixels, item_index):
        full_map_texture = textures["full_map_item_textures"][item_index]
        mask = map_pixels_item_indexes == item_index

        pixels = pixels * (1 - full_map_texture[:, :, 3] * mask[:, :, 0])[:, :, None]
        pixels = (
            pixels
            + full_map_texture[:, :, :3] * mask * full_map_texture[:, :, 3][:, :, None]
        )

        return pixels, None

    map_pixels, _ = jax.lax.scan(
        _add_item_type_to_pixels, map_pixels, jnp.arange(1, len(ItemType))
    )

    # Render player
    player_texture_index = jax.lax.select(
        state.is_sleeping, 4, state.player_direction - 1
    )
    map_pixels = (
        map_pixels
        * (1 - textures["full_map_player_textures_alpha"][player_texture_index])
        + textures["full_map_player_textures"][player_texture_index]
        * textures["full_map_player_textures_alpha"][player_texture_index]
    )

    # Render mobs
    # Zombies

    def _add_mob_to_pixels(carry, mob_index):
        pixels, mobs, texture_name, alpha_texture_name = carry
        local_position = (
            mobs.position[state.player_level, mob_index]
            - state.player_position
            + jnp.ones((2,), dtype=jnp.int32) * (obs_dim_array // 2)
        )
        on_screen = jnp.logical_and(
            local_position >= 0, local_position < obs_dim_array
        ).all()
        on_screen *= mobs.mask[state.player_level, mob_index]

        melee_mob_texture = texture_name[mobs.type_id[state.player_level, mob_index]]
        melee_mob_texture_alpha = alpha_texture_name[
            mobs.type_id[state.player_level, mob_index]
        ]

        melee_mob_texture = melee_mob_texture * on_screen

        melee_mob_texture_with_background = 1 - melee_mob_texture_alpha * on_screen

        melee_mob_texture_with_background = (
            melee_mob_texture_with_background
            * jax.lax.dynamic_slice(
                pixels,
                (
                    local_position[0] * block_pixel_size,
                    local_position[1] * block_pixel_size,
                    0,
                ),
                (block_pixel_size, block_pixel_size, 3),
            )
        )

        melee_mob_texture_with_background = (
            melee_mob_texture_with_background
            + melee_mob_texture * melee_mob_texture_alpha
        )

        pixels = jax.lax.dynamic_update_slice(
            pixels,
            melee_mob_texture_with_background,
            (
                local_position[0] * block_pixel_size,
                local_position[1] * block_pixel_size,
                0,
            ),
        )

        return (pixels, mobs, texture_name, alpha_texture_name), None

    (map_pixels, _, _, _), _ = jax.lax.scan(
        _add_mob_to_pixels,
        (
            map_pixels,
            state.melee_mobs,
            textures["melee_mob_textures"],
            textures["melee_mob_texture_alphas"],
        ),
        jnp.arange(state.melee_mobs.mask.shape[1]),
    )

    (map_pixels, _, _, _), _ = jax.lax.scan(
        _add_mob_to_pixels,
        (
            map_pixels,
            state.passive_mobs,
            textures["passive_mob_textures"],
            textures["passive_mob_texture_alphas"],
        ),
        jnp.arange(state.passive_mobs.mask.shape[1]),
    )

    (map_pixels, _, _, _), _ = jax.lax.scan(
        _add_mob_to_pixels,
        (
            map_pixels,
            state.ranged_mobs,
            textures["ranged_mob_textures"],
            textures["ranged_mob_texture_alphas"],
        ),
        jnp.arange(state.ranged_mobs.mask.shape[1]),
    )

    def _add_projectile_to_pixels(carry, projectile_index):
        pixels, projectiles, projectile_directions = carry
        local_position = (
            projectiles.position[state.player_level, projectile_index]
            - state.player_position
            + jnp.ones((2,), dtype=jnp.int32) * (obs_dim_array // 2)
        )
        on_screen = jnp.logical_and(
            local_position >= 0, local_position < obs_dim_array
        ).all()
        on_screen *= projectiles.mask[state.player_level, projectile_index]

        projectile_texture = textures["projectile_textures"][
            projectiles.type_id[state.player_level, projectile_index]
        ]
        projectile_texture_alpha = textures["projectile_texture_alphas"][
            projectiles.type_id[state.player_level, projectile_index]
        ]

        flipped_projectile_texture = jnp.flip(projectile_texture, axis=0)
        flipped_projectile_texture_alpha = jnp.flip(projectile_texture_alpha, axis=0)
        flip_projectile = jnp.logical_or(
            projectile_directions[state.player_level, projectile_index, 0] > 0,
            projectile_directions[state.player_level, projectile_index, 1] > 0,
        )

        projectile_texture = jax.lax.select(
            flip_projectile,
            flipped_projectile_texture,
            projectile_texture,
        )
        projectile_texture_alpha = jax.lax.select(
            flip_projectile,
            flipped_projectile_texture_alpha,
            projectile_texture_alpha,
        )

        transposed_projectile_texture = jnp.transpose(projectile_texture, (1, 0, 2))
        transposed_projectile_texture_alpha = jnp.transpose(
            projectile_texture_alpha, (1, 0, 2)
        )

        projectile_texture = jax.lax.select(
            projectile_directions[state.player_level, projectile_index, 1] != 0,
            transposed_projectile_texture,
            projectile_texture,
        )
        projectile_texture_alpha = jax.lax.select(
            projectile_directions[state.player_level, projectile_index, 1] != 0,
            transposed_projectile_texture_alpha,
            projectile_texture_alpha,
        )

        projectile_texture = projectile_texture * on_screen
        projectile_texture_with_background = 1 - projectile_texture_alpha * on_screen

        projectile_texture_with_background = (
            projectile_texture_with_background
            * jax.lax.dynamic_slice(
                pixels,
                (
                    local_position[0] * block_pixel_size,
                    local_position[1] * block_pixel_size,
                    0,
                ),
                (block_pixel_size, block_pixel_size, 3),
            )
        )

        projectile_texture_with_background = (
            projectile_texture_with_background
            + projectile_texture * projectile_texture_alpha
        )

        pixels = jax.lax.dynamic_update_slice(
            pixels,
            projectile_texture_with_background,
            (
                local_position[0] * block_pixel_size,
                local_position[1] * block_pixel_size,
                0,
            ),
        )

        return (pixels, projectiles, projectile_directions), None

    (map_pixels, _, _), _ = jax.lax.scan(
        _add_projectile_to_pixels,
        (map_pixels, state.mob_projectiles, state.mob_projectile_directions),
        jnp.arange(state.mob_projectiles.mask.shape[1]),
    )

    (map_pixels, _, _), _ = jax.lax.scan(
        _add_projectile_to_pixels,
        (map_pixels, state.player_projectiles, state.player_projectile_directions),
        jnp.arange(state.player_projectiles.mask.shape[1]),
    )

    # Apply darkness (underground)
    light_map = state.light_map[state.player_level]
    padded_light_map = jnp.pad(
        light_map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=False,
    )

    light_map_view = jax.lax.dynamic_slice(padded_light_map, tl_corner, OBS_DIM)
    light_map_pixels = light_map_view.repeat(block_pixel_size, axis=0).repeat(
        block_pixel_size, axis=1
    )

    map_pixels = (light_map_pixels)[:, :, None] * map_pixels

    # Apply night
    night_pixels = textures["night_texture"]
    daylight = state.light_level
    daylight = jax.lax.select(state.player_level == 0, daylight, 1.0)

    if do_night_noise:
        night_noise = (
            jax.random.uniform(state.state_rng, night_pixels.shape[:2]) * 95 + 32
        )
        night_noise = jnp.expand_dims(night_noise, axis=-1).repeat(3, axis=-1)

        night_intensity = 2 * (0.5 - daylight)
        night_intensity = jnp.maximum(night_intensity, 0.0)
        night_mask = textures["night_noise_intensity_texture"] * night_intensity
        night = (1.0 - night_mask) * map_pixels + night_mask * night_noise

        night = night_pixels * 0.5 + 0.5 * night
        map_pixels = daylight * map_pixels + (1 - daylight) * night
    else:
        night_noise = jnp.ones(night_pixels.shape[:2]) * 64
        night_noise = jnp.expand_dims(night_noise, axis=-1).repeat(3, axis=-1)

        night_intensity = 2 * (0.5 - daylight)
        night_intensity = jnp.maximum(night_intensity, 0.0)
        night_mask = (
            jnp.ones_like(textures["night_noise_intensity_texture"])
            * night_intensity
            * 0.5
        )
        night = (1.0 - night_mask) * map_pixels + night_mask * night_noise

        night = night_pixels * 0.5 + 0.5 * night
        map_pixels = daylight * map_pixels + (1 - daylight) * night
        # map_pixels = daylight * map_pixels
        # night_noise = jnp.ones(night_pixels.shape[:2]) * 64

    # Apply sleep
    sleep_pixels = jnp.zeros_like(map_pixels)
    sleep_level = 1.0 - state.is_sleeping * 0.5
    map_pixels = sleep_level * map_pixels + (1 - sleep_level) * sleep_pixels

    # Render mob map
    # mob_map_pixels = (
    #     jnp.array([[[128, 0, 0]]]).repeat(OBS_DIM[0], axis=0).repeat(OBS_DIM[1], axis=1)
    # )
    # padded_mob_map = jnp.pad(
    #     state.mob_map[state.player_level], MAX_OBS_DIM + 2, constant_values=False
    # )
    # mob_map_view = jax.lax.dynamic_slice(padded_mob_map, tl_corner, OBS_DIM)
    # mob_map_pixels = mob_map_pixels * jnp.expand_dims(mob_map_view, axis=-1)
    # mob_map_pixels = mob_map_pixels.repeat(block_pixel_size, axis=0).repeat(
    #     block_pixel_size, axis=1
    # )
    # map_pixels = map_pixels + mob_map_pixels

    # RENDER INVENTORY
    inv_pixel_left_space = (block_pixel_size - int(0.8 * block_pixel_size)) // 2
    inv_pixel_right_space = (
        block_pixel_size - int(0.8 * block_pixel_size) - inv_pixel_left_space
    )

    inv_pixels = jnp.zeros(
        (INVENTORY_OBS_HEIGHT * block_pixel_size, OBS_DIM[1] * block_pixel_size, 3),
        dtype=jnp.float32,
    )

    number_size = int(block_pixel_size * 0.4)
    number_offset = block_pixel_size - number_size
    number_double_offset = block_pixel_size - 2 * number_size

    def _render_digit(pixels, number, x, y):
        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size + number_offset : (x + 1) * block_pixel_size,
        ].mul(1 - textures["number_textures_alpha"][number])

        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size + number_offset : (x + 1) * block_pixel_size,
        ].add(textures["number_textures"][number])

        return pixels

    def _render_two_digit_number(pixels, number, x, y):
        tens = number // 10
        ones = number % 10

        ones_textures = jax.lax.select(
            number == 0,
            textures["number_textures"],
            textures["number_textures_with_zero"],
        )

        ones_textures_alpha = jax.lax.select(
            number == 0,
            textures["number_textures_alpha"],
            textures["number_textures_alpha_with_zero"],
        )

        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size + number_offset : (x + 1) * block_pixel_size,
        ].mul(1 - ones_textures_alpha[ones])

        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size + number_offset : (x + 1) * block_pixel_size,
        ].add(ones_textures[ones])

        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size
            + number_double_offset : x * block_pixel_size
            + number_offset,
        ].mul(1 - textures["number_textures_alpha"][tens])

        pixels = pixels.at[
            y * block_pixel_size + number_offset : (y + 1) * block_pixel_size,
            x * block_pixel_size
            + number_double_offset : x * block_pixel_size
            + number_offset,
        ].add(textures["number_textures"][tens])

        return pixels

    def _render_icon(pixels, texture, x, y):
        return pixels.at[
            block_pixel_size * y
            + inv_pixel_left_space : block_pixel_size * (y + 1)
            - inv_pixel_right_space,
            block_pixel_size * x
            + inv_pixel_left_space : block_pixel_size * (x + 1)
            - inv_pixel_right_space,
        ].set(texture)

    def _render_icon_with_alpha(pixels, texture, x, y):
        existing_slice = pixels[
            block_pixel_size * y
            + inv_pixel_left_space : block_pixel_size * (y + 1)
            - inv_pixel_right_space,
            block_pixel_size * x
            + inv_pixel_left_space : block_pixel_size * (x + 1)
            - inv_pixel_right_space,
        ]

        new_slice = (
            existing_slice * (1 - texture[:, :, 3][:, :, None])
            + texture[:, :, :3] * texture[:, :, 3][:, :, None]
        )

        return pixels.at[
            block_pixel_size * y
            + inv_pixel_left_space : block_pixel_size * (y + 1)
            - inv_pixel_right_space,
            block_pixel_size * x
            + inv_pixel_left_space : block_pixel_size * (x + 1)
            - inv_pixel_right_space,
        ].set(new_slice)

    # Render player stats
    player_health = jnp.maximum(jnp.floor(state.player_health), 1).astype(int)
    health_texture = jax.lax.select(
        player_health > 0,
        textures["health_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, health_texture, 0, 0)
    inv_pixels = _render_two_digit_number(inv_pixels, player_health, 0, 0)

    hunger_texture = jax.lax.select(
        state.player_food > 0,
        textures["hunger_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, hunger_texture, 1, 0)
    inv_pixels = _render_two_digit_number(inv_pixels, state.player_food, 1, 0)

    thirst_texture = jax.lax.select(
        state.player_drink > 0,
        textures["thirst_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, thirst_texture, 2, 0)
    inv_pixels = _render_two_digit_number(inv_pixels, state.player_drink, 2, 0)

    energy_texture = jax.lax.select(
        state.player_energy > 0,
        textures["energy_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, energy_texture, 3, 0)
    inv_pixels = _render_two_digit_number(inv_pixels, state.player_energy, 3, 0)

    mana_texture = jax.lax.select(
        state.player_mana > 0,
        textures["mana_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, mana_texture, 4, 0)
    inv_pixels = _render_two_digit_number(inv_pixels, state.player_mana, 4, 0)

    # Render inventory

    inv_wood_texture = jax.lax.select(
        state.inventory.wood > 0,
        textures["smaller_block_textures"][BlockType.WOOD.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_wood_texture, 0, 2)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.wood, 0, 2)

    inv_stone_texture = jax.lax.select(
        state.inventory.stone > 0,
        textures["smaller_block_textures"][BlockType.STONE.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_stone_texture, 1, 2)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.stone, 1, 2)

    inv_coal_texture = jax.lax.select(
        state.inventory.coal > 0,
        textures["smaller_block_textures"][BlockType.COAL.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_coal_texture, 0, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.coal, 0, 1)

    inv_iron_texture = jax.lax.select(
        state.inventory.iron > 0,
        textures["smaller_block_textures"][BlockType.IRON.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_iron_texture, 1, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.iron, 1, 1)

    inv_diamond_texture = jax.lax.select(
        state.inventory.diamond > 0,
        textures["smaller_block_textures"][BlockType.DIAMOND.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_diamond_texture, 2, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.diamond, 2, 1)

    inv_sapphire_texture = jax.lax.select(
        state.inventory.sapphire > 0,
        textures["smaller_block_textures"][BlockType.SAPPHIRE.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_sapphire_texture, 3, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.sapphire, 3, 1)

    inv_ruby_texture = jax.lax.select(
        state.inventory.ruby > 0,
        textures["smaller_block_textures"][BlockType.RUBY.value],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_ruby_texture, 4, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.ruby, 4, 1)

    inv_sapling_texture = jax.lax.select(
        state.inventory.sapling > 0,
        textures["sapling_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, inv_sapling_texture, 5, 1)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.sapling, 5, 1)

    # Render tools
    # Pickaxe
    pickaxe_texture = textures["pickaxe_textures"][state.inventory.pickaxe]
    inv_pixels = _render_icon(inv_pixels, pickaxe_texture, 8, 2)

    # Sword
    sword_texture = textures["sword_textures"][state.inventory.sword]
    inv_pixels = _render_icon(inv_pixels, sword_texture, 8, 1)

    # Bow and arrows
    bow_texture = textures["bow_textures"][state.inventory.bow]
    inv_pixels = _render_icon(inv_pixels, bow_texture, 6, 1)

    arrow_texture = jax.lax.select(
        state.inventory.arrows > 0,
        textures["player_projectile_textures"][0],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, arrow_texture, 6, 2)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.arrows, 6, 2)

    # Armour
    for i in range(4):
        armour_texture = textures["armour_textures"][state.inventory.armour[i], i]
        inv_pixels = _render_icon(inv_pixels, armour_texture, 7, i)

    # Torch
    torch_texture = jax.lax.select(
        state.inventory.torches > 0,
        textures["torch_inv_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, torch_texture, 2, 2)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.torches, 2, 2)

    # Potions
    potion_names = ["red", "green", "blue", "pink", "cyan", "yellow"]
    for potion_index, potion_name in enumerate(potion_names):
        potion_texture = jax.lax.select(
            state.inventory.potions[potion_index] > 0,
            textures["potion_textures"][potion_index],
            textures["smaller_empty_texture"],
        )
        inv_pixels = _render_icon(inv_pixels, potion_texture, potion_index, 3)
        inv_pixels = _render_two_digit_number(
            inv_pixels, state.inventory.potions[potion_index], potion_index, 3
        )

    # Books
    book_texture = jax.lax.select(
        state.inventory.books > 0,
        textures["book_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, book_texture, 3, 2)
    inv_pixels = _render_two_digit_number(inv_pixels, state.inventory.books, 3, 2)

    # Learned spells
    fireball_texture = jax.lax.select(
        state.learned_spells[0],
        textures["fireball_inv_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, fireball_texture, 4, 2)

    iceball_texture = jax.lax.select(
        state.learned_spells[1],
        textures["iceball_inv_texture"],
        textures["smaller_empty_texture"],
    )
    inv_pixels = _render_icon(inv_pixels, iceball_texture, 5, 2)

    # Enchantments
    sword_enchantment_texture = textures["sword_enchantment_textures"][
        state.sword_enchantment
    ]
    inv_pixels = _render_icon_with_alpha(inv_pixels, sword_enchantment_texture, 8, 1)

    arrow_enchantment_level = state.bow_enchantment * (state.inventory.arrows > 0)
    arrow_enchantment_texture = textures["arrow_enchantment_textures"][
        arrow_enchantment_level
    ]
    inv_pixels = _render_icon_with_alpha(inv_pixels, arrow_enchantment_texture, 6, 2)

    for i in range(4):
        armour_enchantment_texture = textures["armour_enchantment_textures"][
            state.armour_enchantments[i], i
        ]
        inv_pixels = _render_icon_with_alpha(
            inv_pixels, armour_enchantment_texture, 7, i
        )

    # Dungeon level
    inv_pixels = _render_digit(inv_pixels, state.player_level, 6, 0)

    # Attributes
    xp_texture = jax.lax.select(
        state.player_xp > 0, textures["xp_texture"], textures["smaller_empty_texture"]
    )
    inv_pixels = _render_icon(inv_pixels, xp_texture, 9, 0)
    inv_pixels = _render_digit(inv_pixels, state.player_xp, 9, 0)

    inv_pixels = _render_icon(inv_pixels, textures["dex_texture"], 9, 1)
    inv_pixels = _render_digit(inv_pixels, state.player_dexterity, 9, 1)

    inv_pixels = _render_icon(inv_pixels, textures["str_texture"], 9, 2)
    inv_pixels = _render_digit(inv_pixels, state.player_strength, 9, 2)

    inv_pixels = _render_icon(inv_pixels, textures["int_texture"], 9, 3)
    inv_pixels = _render_digit(inv_pixels, state.player_intelligence, 9, 3)

    # Combine map and inventory
    pixels = jnp.concatenate([map_pixels, inv_pixels], axis=0)

    # # Downscale by 2
    # pixels = pixels[::downscale, ::downscale]

    return pixels


def render_craftax_text(state: EnvState):

    text_obs = "Map:\n"

    # Map
    map = state.map[state.player_level]

    obs_dim_array = jnp.array([OBS_DIM[0], OBS_DIM[1]], dtype=jnp.int32)

    # Map
    padded_grid = jnp.pad(
        map,
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=BlockType.OUT_OF_BOUNDS.value,
    )

    tl_corner = state.player_position - obs_dim_array // 2 + MAX_OBS_DIM + 2

    map_view = jax.lax.dynamic_slice(padded_grid, tl_corner, OBS_DIM)

    padded_items_map = jnp.pad(
        state.item_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=ItemType.NONE.value,
    )

    item_map_view = jax.lax.dynamic_slice(padded_items_map, tl_corner, OBS_DIM)

    # Mobs
    mob_types_per_class = 8
    mob_map = jnp.zeros(
        (*OBS_DIM, 5 * mob_types_per_class), dtype=jnp.int32
    )  # 5 classes * 8 types

    def _add_mob_to_map(carry, mob_index):
        mob_map, mobs, mob_class_index = carry

        local_position = (
            mobs.position[mob_index]
            - state.player_position
            + jnp.array([OBS_DIM[0], OBS_DIM[1]]) // 2
        )
        on_screen = jnp.logical_and(
            local_position >= 0, local_position < jnp.array([OBS_DIM[0], OBS_DIM[1]])
        ).all()
        on_screen *= mobs.mask[mob_index]

        mob_identifier = mob_class_index * mob_types_per_class + mobs.type_id[mob_index]
        mob_map = mob_map.at[local_position[0], local_position[1], mob_identifier].set(
            on_screen.astype(jnp.int32)
        )

        return (mob_map, mobs, mob_class_index), None

    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.melee_mobs), 0),
        jnp.arange(state.melee_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.passive_mobs), 1),
        jnp.arange(state.passive_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (mob_map, jax.tree_map(lambda x: x[state.player_level], state.ranged_mobs), 2),
        jnp.arange(state.ranged_mobs.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (
            mob_map,
            jax.tree_map(lambda x: x[state.player_level], state.mob_projectiles),
            3,
        ),
        jnp.arange(state.mob_projectiles.mask.shape[1]),
    )
    (mob_map, _, _), _ = jax.lax.scan(
        _add_mob_to_map,
        (
            mob_map,
            jax.tree_map(lambda x: x[state.player_level], state.player_projectiles),
            4,
        ),
        jnp.arange(state.player_projectiles.mask.shape[1]),
    )

    def mob_id_to_name(id):
        if id == 0:
            return "Zombie"
        elif id == 1:
            return "Gnome Warrior"
        elif id == 2:
            return "Orc Soldier"
        elif id == 3:
            return "Lizard"
        elif id == 4:
            return "Knight"
        elif id == 5:
            return "Troll"
        elif id == 6:
            return "Pigman"
        elif id == 7:
            return "Frost Troll"
        elif id == 8:
            return "Cow"
        elif id == 9:
            return "Bat"
        elif id == 10:
            return "Snail"
        elif id == 16:
            return "Skeleton"
        elif id == 17:
            return "Gnome Archer"
        elif id == 18:
            return "Orc Mage"
        elif id == 19:
            return "Kobold"
        elif id == 20:
            return "Archer"
        elif id == 21:
            return "Deep Thing"
        elif id == 22:
            return "Fire Elemental"
        elif id == 23:
            return "Ice Elemental"
        elif id == 24:
            return "Arrow"
        elif id == 25:
            return "Dagger"
        elif id == 26:
            return "Fireball"
        elif id == 27:
            return "Iceball"
        elif id == 28:
            return "Arrow"
        elif id == 29:
            return "Slimeball"
        elif id == 30:
            return "Fireball"
        elif id == 31:
            return "Iceball"
        elif id == 32:
            return "Arrow (Player)"
        elif id == 33:
            return "Dagger (Player)"
        elif id == 34:
            return "Fireball (Player)"
        elif id == 35:
            return "Iceball (Player)"
        elif id == 36:
            return "Arrow (Player)"
        elif id == 37:
            return "Slimeball (Player)"
        elif id == 38:
            return "Fireball (Player)"
        elif id == 39:
            return "Iceball (Player)"

    padded_light_map = jnp.pad(
        state.light_map[state.player_level],
        (MAX_OBS_DIM + 2, MAX_OBS_DIM + 2),
        constant_values=0.0,
    )
    light_map_view = jax.lax.dynamic_slice(padded_light_map, tl_corner, OBS_DIM) > 0.05

    for x in range(OBS_DIM[0]):
        for y in range(OBS_DIM[1]):
            text_obs += f"{y - OBS_DIM[1] // 2}, {x - OBS_DIM[0] // 2}: "
            if light_map_view[x, y]:
                if mob_map[x, y].max() > 0.5:
                    text_obs += mob_id_to_name(mob_map[x, y].argmax()) + " on "
                if item_map_view[x, y] != ItemType.NONE.value:
                    text_obs += ItemType(item_map_view[x, y]).name.lower() + " on "
                text_obs += BlockType(map_view[x, y]).name.lower() + "\n"
            else:
                text_obs += "Darkness\n"

    # Inventory
    text_obs += "\nInventory:\n"
    text_obs += f"Wood: {state.inventory.wood}\n"
    text_obs += f"Stone: {state.inventory.stone}\n"
    text_obs += f"Coal: {state.inventory.coal}\n"
    text_obs += f"Iron: {state.inventory.iron}\n"
    text_obs += f"Diamond: {state.inventory.diamond}\n"
    text_obs += f"Sapphire: {state.inventory.sapphire}\n"
    text_obs += f"Ruby: {state.inventory.ruby}\n"
    text_obs += f"Sapling: {state.inventory.sapling}\n"
    text_obs += f"Torch: {state.inventory.torches}\n"
    text_obs += f"Arrow: {state.inventory.arrows}\n"
    text_obs += f"Book: {state.inventory.books}\n"

    def level_to_material(level):
        if level == 1:
            return "Wood"
        elif level == 2:
            return "Stone"
        elif level == 3:
            return "Iron"
        elif level == 4:
            return "Diamond"

    def level_to_enchantment(level):
        if level == 0:
            return "No"
        if level == 1:
            return "Fire"
        elif level == 2:
            return "Ice"

    if state.inventory.pickaxe > 0:
        text_obs += level_to_material(state.inventory.pickaxe) + " Pickaxe\n"
    if state.inventory.sword > 0:
        text_obs += level_to_material(state.inventory.sword) + " Sword"
        text_obs += (
            " with " + level_to_enchantment(state.sword_enchantment) + " enchantment\n"
        )
    if state.inventory.bow > 0:
        text_obs += (
            "Bow with " + level_to_enchantment(state.bow_enchantment) + " enchantment\n"
        )

    text_obs += f"Red potion: {state.inventory.potions[0]}\n"
    text_obs += f"Green potion: {state.inventory.potions[1]}\n"
    text_obs += f"Blue potion: {state.inventory.potions[2]}\n"
    text_obs += f"Pink potion: {state.inventory.potions[3]}\n"
    text_obs += f"Cyan potion: {state.inventory.potions[4]}\n"
    text_obs += f"Yellow potion: {state.inventory.potions[5]}\n"

    def get_armour_level(level):
        if level == 1:
            return "Iron"
        elif level == 2:
            return "Diamond"

    if state.inventory.armour[0] > 0:
        text_obs += f"{get_armour_level(state.inventory.armour[0])} Helmet"
        text_obs += (
            " with "
            + level_to_enchantment(state.armour_enchantments[0])
            + " enchantment\n"
        )

    if state.inventory.armour[1] > 0:
        text_obs += f"{get_armour_level(state.inventory.armour[1])} Chestplate"
        text_obs += (
            " with "
            + level_to_enchantment(state.armour_enchantments[1])
            + " enchantment\n"
        )

    if state.inventory.armour[2] > 0:
        text_obs += f"{get_armour_level(state.inventory.armour[2])} Leggings"
        text_obs += (
            " with "
            + level_to_enchantment(state.armour_enchantments[2])
            + " enchantment\n"
        )

    if state.inventory.armour[3] > 0:
        text_obs += f"{get_armour_level(state.inventory.armour[3])} Boots"
        text_obs += (
            " with "
            + level_to_enchantment(state.armour_enchantments[3])
            + " enchantment\n"
        )

    text_obs += f"Health: {state.player_health}\n"
    text_obs += f"Food: {state.player_food}\n"
    text_obs += f"Drink: {state.player_drink}\n"
    text_obs += f"Energy: {state.player_energy}\n"
    text_obs += f"Mana: {state.player_mana}\n"
    text_obs += f"XP: {state.player_xp}\n"
    text_obs += f"Dexterity: {state.player_dexterity}\n"
    text_obs += f"Strength: {state.player_strength}\n"
    text_obs += f"Intelligence: {state.player_intelligence}\n"

    text_obs += f"Direction: {Action(state.player_direction).name.lower()}\n"

    text_obs += f"Light: {state.light_level}\n"
    text_obs += f"Is Sleeping: {state.is_sleeping}\n"
    text_obs += f"Is Resting: {state.is_resting}\n"
    text_obs += f"Learned Fireball: {state.learned_spells[0]}\n"
    text_obs += f"Learned Iceball: {state.learned_spells[1]}\n"
    text_obs += f"Floor: {state.player_level}\n"
    text_obs += f"Ladder Open: {state.monsters_killed[state.player_level] >= MONSTERS_KILLED_TO_CLEAR_LEVEL}\n"
    text_obs += f"Is Boss Vulnerable: {is_boss_vulnerable(state)}\n"

    return text_obs



from dataclasses import dataclass
from typing import Tuple, Any

import jax.random
from flax import struct
import jax.numpy as jnp


@struct.dataclass
class Inventory:
    wood: int = 0
    stone: int = 0
    coal: int = 0
    iron: int = 0
    diamond: int = 0
    sapling: int = 0
    wood_pickaxe: int = 0
    stone_pickaxe: int = 0
    iron_pickaxe: int = 0
    wood_sword: int = 0
    stone_sword: int = 0
    iron_sword: int = 0


@struct.dataclass
class Mobs:
    position: jnp.ndarray
    health: int
    mask: bool
    attack_cooldown: int


@struct.dataclass
class EnvState:
    map: jnp.ndarray
    mob_map: jnp.ndarray

    player_position: jnp.ndarray
    player_direction: int

    # Intrinsics
    player_health: int
    player_food: int
    player_drink: int
    player_energy: int
    is_sleeping: bool

    # Second order intrinsics
    player_recover: float
    player_hunger: float
    player_thirst: float
    player_fatigue: float

    inventory: Inventory

    zombies: Mobs
    cows: Mobs
    skeletons: Mobs
    arrows: Mobs
    arrow_directions: jnp.ndarray

    growing_plants_positions: jnp.ndarray
    growing_plants_age: jnp.ndarray
    growing_plants_mask: jnp.ndarray

    light_level: float

    achievements: jnp.ndarray

    state_rng: Any

    timestep: int

    fractal_noise_angles: tuple[int, int, int, int] = (None, None, None, None)


@struct.dataclass
class EnvParams:
    max_timesteps: int = 10000
    day_length: int = 300

    always_diamond: bool = True

    zombie_health: int = 5
    cow_health: int = 3
    skeleton_health: int = 3

    mob_despawn_distance: int = 14

    spawn_cow_chance: float = 0.1
    spawn_zombie_base_chance: float = 0.02
    spawn_zombie_night_chance: float = 0.1
    spawn_skeleton_chance: float = 0.05

    fractal_noise_angles: tuple[int, int, int, int] = (None, None, None, None)


@struct.dataclass
class StaticEnvParams:
    map_size: Tuple[int, int] = (64, 64)

    # Mobs
    max_zombies: int = 3
    max_cows: int = 3
    max_growing_plants: int = 10
    max_skeletons: int = 2
    max_arrows: int = 3



from craftax.craftax.envs.craftax_pixels_env import (
    CraftaxPixelsEnv,
    CraftaxPixelsEnvNoAutoReset,
)
from craftax.craftax.envs.craftax_symbolic_env import (
    CraftaxSymbolicEnv,
    CraftaxSymbolicEnvNoAutoReset,
)
from craftax.craftax_classic.envs.craftax_pixels_env import (
    CraftaxClassicPixelsEnv,
    CraftaxClassicPixelsEnvNoAutoReset,
)
from craftax.craftax_classic.envs.craftax_symbolic_env import (
    CraftaxClassicSymbolicEnv,
    CraftaxClassicSymbolicEnvNoAutoReset,
)


def make_craftax_env_from_name(name: str, auto_reset: bool):
    if auto_reset:
        if name == "Craftax-Symbolic-v1" or name == "Craftax-Symbolic-AutoReset-v1":
            return CraftaxSymbolicEnv()
        elif name == "Craftax-Pixels-v1" or name == "Craftax-Pixels-AutoReset-v1":
            return CraftaxPixelsEnv()
        if (
            name == "Craftax-Classic-Symbolic-v1"
            or name == "Craftax-Classic-Symbolic-AutoReset-v1"
        ):
            return CraftaxClassicSymbolicEnv()
        elif (
            name == "Craftax-Classic-Pixels-v1"
            or name == "Craftax-Classic-Pixels-AutoReset-v1"
        ):
            return CraftaxClassicPixelsEnv()
    else:
        if name == "Craftax-Symbolic-v1":
            return CraftaxSymbolicEnvNoAutoReset()
        elif name == "Craftax-Pixels-v1":
            return CraftaxPixelsEnvNoAutoReset()
        elif name == "Craftax-Classic-Symbolic-v1":
            return CraftaxClassicSymbolicEnvNoAutoReset()
        elif name == "Craftax-Classic-Pixels-v1":
            return CraftaxClassicPixelsEnvNoAutoReset()

    raise ValueError(f"Unknown craftax environment: {name}")


def make_craftax_env_from_params(classic: bool, symbolic: bool, auto_reset: bool):
    if classic:
        if symbolic:
            if auto_reset:
                return CraftaxClassicSymbolicEnv()
            else:
                return CraftaxClassicSymbolicEnvNoAutoReset()
        else:
            if auto_reset:
                return CraftaxClassicPixelsEnv()
            else:
                return CraftaxClassicPixelsEnvNoAutoReset()
    else:
        if symbolic:
            if auto_reset:
                return CraftaxSymbolicEnv()
            else:
                return CraftaxSymbolicEnvNoAutoReset()
        else:
            if auto_reset:
                return CraftaxPixelsEnv()
            else:
                return CraftaxPixelsEnvNoAutoReset()



import jax
from flax import struct
import jax.numpy as jnp

from craftax.craftax.constants import BlockType


@struct.dataclass
class SmoothGenConfig:
    default_block: int
    sea_block: int
    coast_block: int
    mountain_block: int
    path_block: int
    inner_mountain_block: int
    ore_requirement_blocks: jnp.ndarray
    ores: jnp.ndarray
    ore_chances: jnp.ndarray
    tree_requirement_block: int
    tree: int
    lava: int
    player_spawn: int
    valid_ladder: int
    ladder_up: int
    ladder_down: int
    player_proximity_map_water_strength: int
    player_proximity_map_water_max: int
    player_proximity_map_mountain_strength: int
    player_proximity_map_mountain_max: int
    default_light: float
    water_threshold: float
    sand_threshold: float
    tree_threshold_uniform: float
    tree_threshold_perlin: float


OVERWORLD_CONFIG = SmoothGenConfig(
    default_block=BlockType.GRASS.value,
    sea_block=BlockType.WATER.value,
    coast_block=BlockType.SAND.value,
    mountain_block=BlockType.STONE.value,
    path_block=BlockType.PATH.value,
    inner_mountain_block=BlockType.PATH.value,
    ore_requirement_blocks=jnp.array([BlockType.STONE.value] * 5),
    ores=jnp.array(
        [
            BlockType.COAL.value,
            BlockType.IRON.value,
            BlockType.DIAMOND.value,
            BlockType.OUT_OF_BOUNDS.value,
            BlockType.OUT_OF_BOUNDS.value,
        ]
    ),
    ore_chances=jnp.array([0.03, 0.02, 0.001, 0.0, 0.0]),
    tree_requirement_block=BlockType.GRASS.value,
    tree=BlockType.TREE.value,
    lava=BlockType.LAVA.value,
    player_spawn=BlockType.GRASS.value,
    valid_ladder=BlockType.PATH.value,
    ladder_up=False,
    ladder_down=True,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=5,
    player_proximity_map_mountain_max=1,
    default_light=1.0,
    water_threshold=0.7,
    sand_threshold=0.6,
    tree_threshold_uniform=0.8,
    tree_threshold_perlin=0.5,
)

GNOMISH_MINES_CONFIG = SmoothGenConfig(
    default_block=BlockType.PATH.value,
    sea_block=BlockType.WATER.value,
    coast_block=BlockType.PATH.value,
    mountain_block=BlockType.STONE.value,
    path_block=BlockType.STONE.value,
    inner_mountain_block=BlockType.STONE.value,
    ore_requirement_blocks=jnp.array([BlockType.STONE.value] * 5),
    ores=jnp.array(
        [
            BlockType.COAL.value,
            BlockType.IRON.value,
            BlockType.DIAMOND.value,
            BlockType.SAPPHIRE.value,
            BlockType.RUBY.value,
        ]
    ),
    ore_chances=jnp.array([0.04, 0.02, 0.005, 0.0025, 0.0025]),
    tree_requirement_block=BlockType.PATH.value,
    tree=BlockType.STALAGMITE.value,
    lava=BlockType.LAVA.value,
    player_spawn=BlockType.PATH.value,
    valid_ladder=BlockType.PATH.value,
    ladder_up=True,
    ladder_down=True,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=17,
    player_proximity_map_mountain_max=1.5,
    default_light=0.0,
    water_threshold=0.7,
    sand_threshold=0.6,
    tree_threshold_uniform=0.8,
    tree_threshold_perlin=0.5,
)

TROLL_MINES_CONFIG = SmoothGenConfig(
    default_block=BlockType.PATH.value,
    sea_block=BlockType.WATER.value,
    coast_block=BlockType.PATH.value,
    mountain_block=BlockType.STONE.value,
    path_block=BlockType.STONE.value,
    inner_mountain_block=BlockType.STONE.value,
    ore_requirement_blocks=jnp.array([BlockType.STONE.value] * 5),
    ores=jnp.array(
        [
            BlockType.COAL.value,
            BlockType.IRON.value,
            BlockType.DIAMOND.value,
            BlockType.SAPPHIRE.value,
            BlockType.RUBY.value,
        ]
    ),
    ore_chances=jnp.array([0.04, 0.03, 0.01, 0.01, 0.01]),
    tree_requirement_block=BlockType.PATH.value,
    tree=BlockType.STALAGMITE.value,
    lava=BlockType.LAVA.value,
    player_spawn=BlockType.PATH.value,
    valid_ladder=BlockType.PATH.value,
    ladder_up=True,
    ladder_down=True,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=17,
    player_proximity_map_mountain_max=1.5,
    default_light=0.0,
    water_threshold=0.7,
    sand_threshold=0.6,
    tree_threshold_uniform=0.8,
    tree_threshold_perlin=0.5,
)

FIRE_LEVEL_CONFIG = SmoothGenConfig(
    default_block=BlockType.FIRE_GRASS.value,
    sea_block=BlockType.LAVA.value,
    coast_block=BlockType.SAND.value,
    mountain_block=BlockType.STONE.value,
    path_block=BlockType.STONE.value,
    inner_mountain_block=BlockType.STONE.value,
    ore_requirement_blocks=jnp.array([BlockType.STONE.value] * 5),
    ores=jnp.array(
        [
            BlockType.COAL.value,
            BlockType.IRON.value,
            BlockType.DIAMOND.value,
            BlockType.SAPPHIRE.value,
            BlockType.RUBY.value,
        ]
    ),
    ore_chances=jnp.array([0.05, 0.0, 0.0, 0.0, 0.025]),
    tree_requirement_block=BlockType.FIRE_GRASS.value,
    tree=BlockType.FIRE_TREE.value,
    lava=BlockType.LAVA.value,
    player_spawn=BlockType.FIRE_GRASS.value,
    valid_ladder=BlockType.FIRE_GRASS.value,
    ladder_up=True,
    ladder_down=True,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=5,
    player_proximity_map_mountain_max=1,
    default_light=1.0,
    water_threshold=0.5,
    sand_threshold=0.6,
    tree_threshold_uniform=0.8,
    tree_threshold_perlin=0.5,
)

ICE_LEVEL_CONFIG = SmoothGenConfig(
    default_block=BlockType.ICE_GRASS.value,
    sea_block=BlockType.WATER.value,
    coast_block=BlockType.ICE_GRASS.value,
    mountain_block=BlockType.STONE.value,
    path_block=BlockType.STONE.value,
    inner_mountain_block=BlockType.STONE.value,
    ore_requirement_blocks=jnp.array([BlockType.STONE.value] * 5),
    ores=jnp.array(
        [
            BlockType.COAL.value,
            BlockType.IRON.value,
            BlockType.DIAMOND.value,
            BlockType.SAPPHIRE.value,
            BlockType.RUBY.value,
        ]
    ),
    ore_chances=jnp.array([0.0, 0.0, 0.005, 0.02, 0.0]),
    tree_requirement_block=BlockType.ICE_GRASS.value,
    tree=BlockType.ICE_SHRUB.value,
    lava=BlockType.WATER.value,
    player_spawn=BlockType.ICE_GRASS.value,
    valid_ladder=BlockType.ICE_GRASS.value,
    ladder_up=True,
    ladder_down=True,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=17,
    player_proximity_map_mountain_max=1.5,
    default_light=0.0,
    water_threshold=0.5,
    sand_threshold=0.6,
    tree_threshold_uniform=0.4,
    tree_threshold_perlin=0.5,
)

BOSS_LEVEL_CONFIG = SmoothGenConfig(
    default_block=BlockType.PATH.value,
    sea_block=BlockType.PATH.value,
    coast_block=BlockType.PATH.value,
    mountain_block=BlockType.WALL.value,
    path_block=BlockType.WALL.value,
    inner_mountain_block=BlockType.WALL.value,
    ore_requirement_blocks=jnp.array(
        [
            BlockType.WALL.value,
            BlockType.GRAVE.value,
            BlockType.GRAVE.value,
            BlockType.WALL.value,
            BlockType.WALL.value,
        ]
    ),
    ores=jnp.array(
        [
            BlockType.WALL_MOSS.value,
            BlockType.GRAVE2.value,
            BlockType.GRAVE3.value,
            BlockType.SAPPHIRE.value,
            BlockType.RUBY.value,
        ]
    ),
    ore_chances=jnp.array([0.1, 0.333, 0.5, 0.0, 0.0]),
    tree_requirement_block=BlockType.PATH.value,
    tree=BlockType.GRAVE.value,
    lava=BlockType.WALL.value,
    player_spawn=BlockType.NECROMANCER.value,
    valid_ladder=BlockType.PATH.value,
    ladder_up=False,
    ladder_down=False,
    player_proximity_map_water_strength=5,
    player_proximity_map_water_max=1,
    player_proximity_map_mountain_strength=10,
    player_proximity_map_mountain_max=10,
    default_light=0.0,
    water_threshold=0.7,
    sand_threshold=0.6,
    tree_threshold_uniform=0.95,
    tree_threshold_perlin=-1.0,
)

ALL_SMOOTHGEN_CONFIGS = jax.tree_map(
    lambda l1, l2, l3, l4, l5, l6: jnp.stack((l1, l2, l3, l4, l5, l6), axis=0),
    OVERWORLD_CONFIG,
    GNOMISH_MINES_CONFIG,
    TROLL_MINES_CONFIG,
    FIRE_LEVEL_CONFIG,
    ICE_LEVEL_CONFIG,
    BOSS_LEVEL_CONFIG,
)


@struct.dataclass
class DungeonConfig:
    special_block: int
    fountain_block: int
    rare_path_replacement_block: int


DUNGEON_CONFIG = DungeonConfig(
    special_block=BlockType.PATH.value,
    fountain_block=BlockType.FOUNTAIN.value,
    rare_path_replacement_block=BlockType.PATH.value,
)

SEWER_CONFIG = DungeonConfig(
    special_block=BlockType.ENCHANTMENT_TABLE_ICE.value,
    fountain_block=BlockType.WATER.value,
    rare_path_replacement_block=BlockType.WATER.value,
)

VAULTS_CONFIG = DungeonConfig(
    special_block=BlockType.ENCHANTMENT_TABLE_FIRE.value,
    fountain_block=BlockType.FOUNTAIN.value,
    rare_path_replacement_block=BlockType.PATH.value,
)

ALL_DUNGEON_CONFIGS = jax.tree_map(
    lambda x, y, z: jnp.stack((x, y, z), axis=0),
    DUNGEON_CONFIG,
    SEWER_CONFIG,
    VAULTS_CONFIG,
)



import pickle
import bz2
from typing import Any


def save_compressed_pickle(title: str, data: Any):
    with bz2.BZ2File(title, "w") as f:
        pickle.dump(data, f)


def load_compressed_pickle(file: str):
    data = bz2.BZ2File(file, "rb")
    data = pickle.load(data)
    return data



import chex

from craftax.craftax.constants import *
from craftax.craftax.craftax_state import *

# For utility functions - functions called more than once in meaningfully different parts of the codebase


def is_fighting_boss(state, static_params):
    return state.player_level == (static_params.num_levels - 1)


def is_boss_spawn_wave(state, static_params):
    return jnp.logical_and(
        is_fighting_boss(state, static_params),
        state.boss_timesteps_to_spawn_this_round >= 1,
    )


def is_boss_vulnerable(state):
    return jnp.logical_and(
        state.melee_mobs.mask[state.player_level].sum() == 0,
        jnp.logical_and(
            state.ranged_mobs.mask[state.player_level].sum() == 0,
            state.boss_timesteps_to_spawn_this_round <= 0,
        ),
    )


def has_beaten_boss(state, static_params):
    return state.boss_progress >= static_params.num_levels - 1


def attack_mob_class(
    state,
    mobs,
    position,
    damage_vector,
    can_get_achievement,
    mob_class_index,
):
    def is_attacking_mob_at_index(unused, mob_index):
        in_mob = (mobs.position[state.player_level, mob_index] == position).all()
        return None, jnp.logical_and(in_mob, mobs.mask[state.player_level, mob_index])

    _, is_attacking_mob_array = jax.lax.scan(
        is_attacking_mob_at_index, None, jnp.arange(mobs.mask.shape[1])
    )
    is_attacking_mob = is_attacking_mob_array.sum() > 0
    target_mob_index = jnp.argmax(is_attacking_mob_array)

    damage = get_damage(
        damage_vector,
        MOB_TYPE_DEFENSE_MAPPING[
            mobs.type_id[state.player_level, target_mob_index], mob_class_index
        ],
    )

    new_mob_health = mobs.health.at[state.player_level, target_mob_index].add(
        -damage * is_attacking_mob
    )
    mobs = mobs.replace(health=new_mob_health)

    old_mask = mobs.mask[state.player_level, target_mob_index]
    mobs = mobs.replace(mask=jnp.logical_and(mobs.health > 0, mobs.mask))
    did_kill_mob = jnp.logical_and(
        old_mask,
        jnp.logical_not(mobs.mask[state.player_level, target_mob_index]),
    )

    achievement_for_kill = MOB_ACHIEVEMENT_MAP[
        mob_class_index, mobs.type_id[state.player_level, target_mob_index]
    ]

    new_achievements = state.achievements.at[achievement_for_kill].set(
        jnp.logical_or(
            state.achievements[achievement_for_kill],
            jnp.logical_and(did_kill_mob, can_get_achievement),
        )
    )

    return mobs, did_kill_mob, is_attacking_mob, new_achievements


def attack_mob(state, position, damage_vector, can_eat):
    # Melee
    (
        new_melee_mobs,
        did_kill_melee_mob,
        is_attacking_melee_mob,
        new_achievements,
    ) = attack_mob_class(
        state,
        state.melee_mobs,
        position,
        damage_vector,
        True,
        1,
    )

    state = state.replace(
        melee_mobs=new_melee_mobs,
        achievements=new_achievements,
    )

    # Cow
    (
        new_passive_mobs,
        did_kill_passive_mob,
        is_attacking_passive_mob,
        new_achievements,
    ) = attack_mob_class(
        state,
        state.passive_mobs,
        position,
        damage_vector,
        can_eat,
        0,
    )

    new_food = jax.lax.select(
        jnp.logical_and(did_kill_passive_mob, can_eat),
        jnp.minimum(get_max_food(state), state.player_food + 6),
        state.player_food,
    )
    new_hunger = jax.lax.select(
        jnp.logical_and(did_kill_passive_mob, can_eat), 0.0, state.player_hunger
    )

    state = state.replace(
        passive_mobs=new_passive_mobs,
        player_food=new_food,
        player_hunger=new_hunger,
        achievements=new_achievements,
    )

    # Skeleton
    (
        new_ranged_mobs,
        did_kill_ranged_mob,
        is_attacking_ranged_mob,
        new_achievements,
    ) = attack_mob_class(
        state,
        state.ranged_mobs,
        position,
        damage_vector,
        True,
        2,
    )

    state = state.replace(
        ranged_mobs=new_ranged_mobs,
        achievements=new_achievements,
    )

    # Update mob map on kill

    did_attack_mob = jnp.logical_or(
        jnp.logical_or(is_attacking_melee_mob, is_attacking_passive_mob),
        is_attacking_ranged_mob,
    )

    did_kill_monster = jnp.logical_or(did_kill_melee_mob, did_kill_ranged_mob)
    did_kill_mob = jnp.logical_or(did_kill_monster, did_kill_passive_mob)

    state = state.replace(
        mob_map=state.mob_map.at[state.player_level, position[0], position[1]].set(
            jnp.logical_and(
                state.mob_map[state.player_level, position[0], position[1]],
                jnp.logical_not(did_kill_mob),
            )
        ),
        monsters_killed=state.monsters_killed.at[state.player_level].add(
            1 * did_kill_monster
        ),
    )

    return state, did_attack_mob, did_kill_mob


def spawn_projectile(
    state,
    static_params,
    projectiles,
    projectile_directions,
    new_projectile_position,
    is_spawning_projectile,
    direction,
    projectile_type,
):
    new_projectile_index = jnp.argmax(
        jnp.logical_not(projectiles.mask[state.player_level])
    )
    new_projectile_position = jax.lax.select(
        is_spawning_projectile,
        new_projectile_position,
        projectiles.position[state.player_level, new_projectile_index],
    )
    new_projectile_mask = jax.lax.select(
        is_spawning_projectile,
        True,
        projectiles.mask[state.player_level, new_projectile_index],
    )
    new_projectile_direction = jax.lax.select(
        is_spawning_projectile,
        direction,
        projectile_directions[state.player_level, new_projectile_index],
    )
    new_projectile_type = jax.lax.select(
        is_spawning_projectile,
        projectile_type,
        projectiles.type_id[state.player_level, new_projectile_index],
    )

    new_projectiles = projectiles.replace(
        position=projectiles.position.at[state.player_level, new_projectile_index].set(
            new_projectile_position
        ),
        mask=projectiles.mask.at[state.player_level, new_projectile_index].set(
            new_projectile_mask
        ),
        type_id=projectiles.type_id.at[state.player_level, new_projectile_index].set(
            new_projectile_type
        ),
    )

    new_projectile_directions = projectile_directions.at[
        state.player_level, new_projectile_index
    ].set(new_projectile_direction)

    return new_projectiles, new_projectile_directions


def get_damage_done_to_player(state, static_params, damage_vector):
    scaled_defenses = jnp.stack(
        [
            state.inventory.armour * 0.1,
            (state.armour_enchantments == 1) * 0.2,
            (state.armour_enchantments == 2) * 0.2,
        ],
        axis=0,
    )

    defense_vector = scaled_defenses.sum(axis=1)

    damage_vector *= (
        1 + is_fighting_boss(state, static_params) * BOSS_FIGHT_EXTRA_DAMAGE
    )

    return get_damage(damage_vector, defense_vector)


def get_player_damage_vector(state):
    physical_damages = jnp.array(
        [1, 2, 3, 5, 8],
        dtype=jnp.int32,
    )
    physical_damage = physical_damages[state.inventory.sword]
    fire_damage = physical_damage * (state.sword_enchantment == 1) * 0.5
    ice_damage = physical_damage * (state.sword_enchantment == 2) * 0.5

    physical_damage *= 1 + 0.25 * (
        state.player_strength - 1
    )  # Strength=5 does double damage
    fire_damage *= 1 + 0.05 * (
        state.player_intelligence - 1
    )  # Int=5 does 25% more enchant damage
    ice_damage *= 1 + 0.05 * (
        state.player_intelligence - 1
    )  # Int=5 does 25% more enchant damage

    return jnp.stack([physical_damage, fire_damage, ice_damage], axis=0)


def get_damage(damage_vector, defense_vector):
    damages = (1.0 - defense_vector) * damage_vector

    return damages.sum()


def in_bounds(state, position):
    in_bounds_x = jnp.logical_and(
        0 <= position[0], position[0] < state.map[state.player_level].shape[0]
    )
    in_bounds_y = jnp.logical_and(
        0 <= position[1], position[1] < state.map[state.player_level].shape[1]
    )
    return jnp.logical_and(in_bounds_x, in_bounds_y)


def is_in_solid_block(state, position):
    return SOLID_BLOCK_MAPPING[state.map[state.player_level, position[0], position[1]]]


def is_position_in_bounds_not_in_mob_not_colliding(state, position, collision_map):
    pos_in_bounds = in_bounds(state, position)
    in_solid_block = is_in_solid_block(state, position)
    in_mob = is_in_mob(state, position)
    in_lava = (
        state.map[state.player_level][position[0], position[1]] == BlockType.LAVA.value
    )
    in_water = (
        state.map[state.player_level][position[0], position[1]] == BlockType.WATER.value
    )
    on_ground_block = jnp.logical_and(
        jnp.logical_not(in_solid_block),
        jnp.logical_and(jnp.logical_not(in_water), jnp.logical_not(in_lava)),
    )

    valid_move = jnp.logical_and(
        pos_in_bounds,
        jnp.logical_and(jnp.logical_not(in_mob), jnp.logical_not(in_solid_block)),
    )

    # Ground blocks
    valid_move = jnp.logical_and(
        valid_move,
        jnp.logical_or(
            jnp.logical_not(collision_map[0]), jnp.logical_not(on_ground_block)
        ),
    )

    # Water
    valid_move = jnp.logical_and(
        valid_move,
        jnp.logical_or(jnp.logical_not(collision_map[1]), jnp.logical_not(in_water)),
    )

    # Lava
    valid_move = jnp.logical_and(
        valid_move,
        jnp.logical_or(jnp.logical_not(collision_map[2]), jnp.logical_not(in_lava)),
    )

    return valid_move


def is_near_block(state, block_type):
    def _is_given_block(unused, loc_add):
        pos = state.player_position + loc_add
        is_in_bounds = in_bounds(state, pos)
        is_correct_block = state.map[state.player_level][pos[0], pos[1]] == block_type
        return None, jnp.logical_and(is_in_bounds, is_correct_block)

    _, is_block = jax.lax.scan(_is_given_block, None, CLOSE_BLOCKS)

    return is_block.sum() > 0


def calculate_light_level(timestep, params):
    progress = (timestep / params.day_length) % 1 + 0.3
    return 1 - jnp.abs(jnp.cos(jnp.pi * progress)) ** 3


def is_in_mob(state: EnvState, position: chex.Array):
    return jnp.logical_or(
        state.mob_map[state.player_level, position[0], position[1]],
        (state.player_position == position).all(),
    )


def get_max_health(state):
    return 8 + state.player_strength


def get_max_food(state):
    return 7 + 2 * state.player_dexterity


def get_max_drink(state):
    return 7 + 2 * state.player_dexterity


def get_max_energy(state):
    return 7 + 2 * state.player_dexterity


def get_max_mana(state):
    return 6 + 3 * state.player_intelligence


def clip_inventory_and_intrinsics(state, params):
    capped_inv = jax.tree_map(lambda x: jnp.minimum(x, 99), state.inventory)

    min_health = jax.lax.select(params.god_mode, 9, 0)

    state = state.replace(
        inventory=capped_inv,
        player_health=jnp.minimum(
            jnp.maximum(state.player_health, min_health), get_max_health(state)
        ),
        player_food=jnp.minimum(jnp.maximum(state.player_food, 0), get_max_food(state)),
        player_drink=jnp.minimum(
            jnp.maximum(state.player_drink, 0), get_max_drink(state)
        ),
        player_energy=jnp.minimum(
            jnp.maximum(state.player_energy, 0), get_max_energy(state)
        ),
        player_mana=jnp.minimum(jnp.maximum(state.player_mana, 0), get_max_mana(state)),
    )

    return state



import jax
import jax.numpy as jnp


def get_distance_map(position, map_size):
    dist_x = jnp.abs(jnp.arange(0, map_size[0]) - position[0])
    dist_x = jnp.expand_dims(dist_x, axis=1)
    dist_x = jnp.tile(dist_x, (1, map_size[1]))

    dist_y = jnp.abs(jnp.arange(0, map_size[1]) - position[1])
    dist_y = jnp.expand_dims(dist_y, axis=0)
    dist_y = jnp.tile(dist_y, (map_size[0], 1))

    coords = jnp.stack([dist_x, dist_y], axis=-1)

    def _euclid_distance(x):
        return jnp.sqrt(x[0] ** 2 + x[1] ** 2)

    dist = jax.vmap(jax.vmap(_euclid_distance))(coords)

    return dist
